#ifndef _MY_QUEUE_H
#define _MY_QUEUE_H

#include <stdatomic.h>

#include <spec_lib.h>
#include <stdlib.h>
#include <cdsannotate.h>
#include <specannotation.h>
#include <model_memory.h>
#include "common.h" 

#define MAX_NODES			0xf

typedef unsigned long long pointer;
typedef atomic_ullong pointer_t;

#define MAKE_POINTER(ptr, count)	((((pointer)count) << 32) | ptr)
#define PTR_MASK 0xffffffffLL
#define COUNT_MASK (0xffffffffLL << 32)

static inline void set_count(pointer *p, unsigned int val) { *p = (*p & ~COUNT_MASK) | ((pointer)val << 32); }
static inline void set_ptr(pointer *p, unsigned int val) { *p = (*p & ~PTR_MASK) | val; }
static inline unsigned int get_count(pointer p) { return (p & COUNT_MASK) >> 32; }
static inline unsigned int get_ptr(pointer p) { return p & PTR_MASK; }

typedef struct node {
	unsigned int value;
	pointer_t next;
} node_t;

typedef struct {
	pointer_t head;
	pointer_t tail;
	node_t nodes[MAX_NODES + 1];
} queue_t;

void init_queue(queue_t *q, int num_threads);

/* All other user-defined structs */
static IntegerList * __queue;
/* All other user-defined functions */
/* Definition of interface info struct: Dequeue */
typedef struct Dequeue_info {
bool __RET__;
queue_t * q;
int * retVal;
} Dequeue_info;
/* End of info struct definition: Dequeue */

/* ID function of interface: Dequeue */
inline static call_id_t Dequeue_id(void *info, thread_id_t __TID__) {
	Dequeue_info* theInfo = (Dequeue_info*)info;
	bool __RET__ = theInfo->__RET__;
	queue_t * q = theInfo->q;
	int * retVal = theInfo->retVal;

	call_id_t __ID__ = __RET__ ? * retVal : 0;
	return __ID__;
}
/* End of ID function: Dequeue */

/* Check action function of interface: Dequeue */
inline static bool Dequeue_check_action(void *info, call_id_t __ID__, thread_id_t __TID__) {
	bool check_passed;
	Dequeue_info* theInfo = (Dequeue_info*)info;
	bool __RET__ = theInfo->__RET__;
	queue_t * q = theInfo->q;
	int * retVal = theInfo->retVal;

	int elem = 0 ;
	if ( __RET__ ) {
	elem = front ( __queue ) ;
	pop_front ( __queue ) ;
	}
	check_passed = __RET__ ? * retVal == elem : true;
	if (!check_passed)
		return false;
	return true;
}
/* End of check action function: Dequeue */

/* Definition of interface info struct: Enqueue */
typedef struct Enqueue_info {
queue_t * q;
 unsigned int val;
} Enqueue_info;
/* End of info struct definition: Enqueue */

/* ID function of interface: Enqueue */
inline static call_id_t Enqueue_id(void *info, thread_id_t __TID__) {
	Enqueue_info* theInfo = (Enqueue_info*)info;
	queue_t * q = theInfo->q;
	 unsigned int val = theInfo->val;

	call_id_t __ID__ = val;
	return __ID__;
}
/* End of ID function: Enqueue */

/* Check action function of interface: Enqueue */
inline static bool Enqueue_check_action(void *info, call_id_t __ID__, thread_id_t __TID__) {
	bool check_passed;
	Enqueue_info* theInfo = (Enqueue_info*)info;
	queue_t * q = theInfo->q;
	 unsigned int val = theInfo->val;

	push_back ( __queue , val ) ;
	return true;
}
/* End of check action function: Enqueue */

#define INTERFACE_SIZE 2
static void** func_ptr_table;
static hb_rule** hb_rule_table;
static commutativity_rule** commutativity_rule_table;
inline static bool CommutativityCondition0(void *info1, void *info2) {
	Enqueue_info *_info1 = (Enqueue_info*) info1;
	Dequeue_info *_info2 = (Dequeue_info*) info2;
	return true;
}
inline static bool CommutativityCondition1(void *info1, void *info2) {
	Dequeue_info *_info1 = (Dequeue_info*) info1;
	Dequeue_info *_info2 = (Dequeue_info*) info2;
	return ! _info1-> __RET__ || ! _info2-> __RET__;
}

/* Initialization of sequential varialbes */
static void __SPEC_INIT__() {
	__queue = createIntegerList ( ) ;
}

/* Cleanup routine of sequential variables */
static bool __SPEC_CLEANUP__() {
	if ( __queue ) destroyIntegerList ( __queue ) ;
	return true ;
}

/* Define function for sequential code initialization */
inline static void __sequential_init() {
	/* Init func_ptr_table */
	func_ptr_table = (void**) malloc(sizeof(void*) * 2 * 2);
	func_ptr_table[2 * 1] = (void*) &Dequeue_id;
	func_ptr_table[2 * 1 + 1] = (void*) &Dequeue_check_action;
	func_ptr_table[2 * 0] = (void*) &Enqueue_id;
	func_ptr_table[2 * 0 + 1] = (void*) &Enqueue_check_action;
	/* Enqueue(true) -> Dequeue(true) */
	struct hb_rule *hbConditionInit0 = (struct hb_rule*) malloc(sizeof(struct hb_rule));
	hbConditionInit0->interface_num_before = 0; // Enqueue
	hbConditionInit0->hb_condition_num_before = 0; // 
	hbConditionInit0->interface_num_after = 1; // Dequeue
	hbConditionInit0->hb_condition_num_after = 0; // 
	/* Init hb_rule_table */
	hb_rule_table = (hb_rule**) malloc(sizeof(hb_rule*) * 1);
	#define HB_RULE_TABLE_SIZE 1
	hb_rule_table[0] = hbConditionInit0;
	/* Init commutativity_rule_table */
	commutativity_rule_table = (commutativity_rule**) malloc(sizeof(commutativity_rule*) * 2);
	commutativity_rule* rule;
	rule = (commutativity_rule*) malloc (sizeof(commutativity_rule));
	rule->interface_num_before = 0;
	rule->interface_num_after = 1;
	rule->rule = "true";
	rule->condition = CommutativityCondition0;
	commutativity_rule_table[0] = rule;
	rule = (commutativity_rule*) malloc (sizeof(commutativity_rule));
	rule->interface_num_before = 1;
	rule->interface_num_after = 1;
	rule->rule = "! _Method1 . __RET__ || ! _Method2 . __RET__";
	rule->condition = CommutativityCondition1;
	commutativity_rule_table[1] = rule;
	/* Pass init info, including function table info & HB rules & Commutativity Rules */
	struct anno_init *anno_init = (struct anno_init*) malloc(sizeof(struct anno_init));
	anno_init->init_func = (void_func_t) __SPEC_INIT__;
	anno_init->cleanup_func = (cleanup_func_t) __SPEC_CLEANUP__;
	anno_init->func_table = func_ptr_table;
	anno_init->func_table_size = INTERFACE_SIZE;
	anno_init->hb_rule_table = hb_rule_table;
	anno_init->hb_rule_table_size = HB_RULE_TABLE_SIZE;
	anno_init->commutativity_rule_table = commutativity_rule_table;
	anno_init->commutativity_rule_table_size = 2;
	struct spec_annotation *init = (struct spec_annotation*) malloc(sizeof(struct spec_annotation));
	init->type = INIT;
	init->annotation = anno_init;
	cdsannotate(SPEC_ANALYSIS, init);

}

/* End of Global construct generation in class */




void __wrapper__enqueue(queue_t * q,  unsigned int val);

void __wrapper__enqueue(queue_t * q,  unsigned int val) ;

bool __wrapper__dequeue(queue_t * q, int * retVal);

bool __wrapper__dequeue(queue_t * q, int * retVal) ;
int get_thread_num();

#endif

