package edu.uci.eecs.specCompiler.specExtraction;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.LineNumberReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;

import edu.uci.eecs.specCompiler.grammerParser.ParseException;
import edu.uci.eecs.specCompiler.grammerParser.SpecParser;
import edu.uci.eecs.specCompiler.grammerParser.TokenMgrError;

/**
 * <p>
 * This class represents the specification extractor of the specification. The
 * main function of this class is to read C/C++11 source files and extract the
 * corresponding specifications, and record corresponding information such as
 * location, e.g., the file name and the line number, to help the code
 * generation process.
 * </p>
 * 
 * @author peizhaoo
 * 
 */
public class SpecExtractor {
	public final HashMap<File, SourceFileInfo> srcFilesInfo;

	public SpecExtractor() {
		srcFilesInfo = new HashMap<File, SourceFileInfo>();
	}

	public ArrayList<Construct> getConstructs() {
		ArrayList<Construct> constructs = new ArrayList<Construct>();
		for (File f : srcFilesInfo.keySet()) {
			constructs.addAll(srcFilesInfo.get(f).constructs);
		}
		return constructs;
	}

	/**
	 * <p>
	 * Given a list of files, it scans each file and add found SpecConstrcut to
	 * the _constructs list.
	 * </p>
	 * 
	 * @param files
	 * @throws SpecNotMatchException
	 */
	public void extract(File[] files) {
		for (int i = 0; i < files.length; i++)
			extract(files[i]);
	}

	public void extract(File file) {
		if (srcFilesInfo.containsKey(file))
			return;
		SourceFileInfo srcFileInfo;
		try {
			srcFileInfo = SpecParser.ParseFile(file);
			for (int i = 0; i < srcFileInfo.content.size(); i++) {
				System.out.println(srcFileInfo.content.get(i));
			}
			ParserUtils.write2File(srcFileInfo.file, srcFileInfo.content);
			srcFilesInfo.put(file, srcFileInfo);
		} catch (ParseException e) {
			e.printStackTrace();
		} catch (TokenMgrError e) {
			e.printStackTrace();
		}
	}

}
